#ifndef UTILITY_H
#define UTILITY_H

#include <stdio.h>
#include <stdlib.h>
#include <vector>

typedef unsigned long uint32;
typedef unsigned short uint16;
typedef unsigned char uint8;
typedef short int16;
typedef unsigned int uint;

// Endian-safe int readers/writers:

uint8
read_uint8 (FILE * f)
{
	return uint8 (fgetc (f));
}

uint32
read_uint32 (FILE * f)
{
	uint32 result = read_uint8 (f);
	result += uint32 (read_uint8 (f)) << 8;
	result += uint32 (read_uint8 (f)) << 16;
	result += uint32 (read_uint8 (f)) << 24;

	return result;
}

uint16
read_uint16 (FILE * f)
{
	uint16 result = read_uint8 (f);
	result += uint16 (read_uint8 (f)) << 8;

	return result;
}

// Number of bits required to store a value:

int
bits_to_store (uint32 i)
{
	int result = 0;
	while (i) {
		result++;
		i >>= 1;
	}
	return result;
}

// Bitstream reader and writer:

struct bitstream_in {
	FILE *f_in;
	int byte;
	int bit_no;

	void setup (FILE * _f_in) {
		f_in = _f_in;
		bit_no = 8;
	} 

	int read () {
		if (bit_no == 8) {
			byte = fgetc (f_in);
			if (byte == EOF)
				return -1;

			bit_no = 0;
		}

		return (byte & (1 << bit_no++) ? 1 : 0);
	}

	uint32 read_uint (int bits) {
		uint32 value = 0;
		for (int i = 0; i < bits; i++)
			value += read () << i;
		return value;
	}

	uint read_uint_max (int max) {
		if (!max)
			return 0;
		int bits = bits_to_store (max);

		uint value = 0;
		for (int i = 0; i < bits - 1; i++)
			if (read ())
				value += 1 << i;

		if ((value | (1 << (bits - 1))) <= max)
			if (read ())
				value += 1 << (bits - 1);

		return value;
	}
};

const int adapt_level = 8;

void
read_list (int *list, int list_size, bool base_2_part, bitstream_in & in)
{
	//Read a list of ints as encoded by "write_list"

	//Read base 2 part:

	int low_bits = (base_2_part ? in.read_uint (4) : 0);

	for (int i = 0; i < list_size; i++)
		list[i] = in.read_uint (low_bits);

	// Decode bitstream:

	int n_zeros = 0;
	int step = 256;
	bool dominant = false;
	vector < uint8 > bits;
	while (n_zeros < list_size) {
		int steplet = step >> 8;

		if (!in.read ()) {
			for (int i = 0; i < steplet; i++)
				bits.push_back (dominant);

			if (!dominant)
				n_zeros += steplet;

			step += step / adapt_level;
		} else {
			int actual_run = in.read_uint_max (steplet - 1);

			for (int i = 0; i < actual_run; i++)
				bits.push_back (dominant);
			bits.push_back (!dominant);

			if (!dominant)
				n_zeros += actual_run;
			if (dominant)
				n_zeros++;

			step -= step / adapt_level;
		}

		if (step < 256) {
			step = 65536 / step;
			dominant = !dominant;
		}
	}

	// Reconstruct (unsigned) values:

	n_zeros = 0;
	int pos = 0;
	int level = 0;
	for (int i = 0; n_zeros < list_size; i++) {
		for (;;) {
			if (pos >= list_size) {
				pos = 0;
				level += 1 << low_bits;
			}

			if (list[pos] >= level)
				break;

			pos++;
		}

		if (bits[i])
			list[pos] += 1 << low_bits;
		else
			n_zeros++;

		pos++;
	}

	// Read signs:

	for (int i = 0; i < list_size; i++)
		if (list[i] && in.read ())
			list[i] = -list[i];
}

#endif				//UTILITY_H
