/* Bonk - lossy/lossless audio compressor
   Copyright (C) 2001  Paul Francis Harrison

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  675 Mass Ave, Cambridge, MA 02139, USA.

  The author may be contacted at:
    pfh@csse.monash.edu.au
  or
    3 Currajong St., Oakleigh East, 3166, Melbourne, Australia

  See also
    http://yoyo.cc.monash.edu.au/~pfh/

  NOTE: This file modified by tainio to suite playbonk.
*/

#include <sys/types.h>
#include <signal.h>

const char *version = "0.5";

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <vector>
#include <string>
#include <algorithm>
#include <unistd.h>

#include "pb-bonk.h"

// Support from Linux and *BSD sound output

#if defined(__linux__)
#  define dsp_device "/dev/dsp"
#  include <linux/soundcard.h>
#elif defined(__FreeBSD__)
#  define dsp_device "/dev/dsp"
#  include <sys/soundcard.h>
#  if !defined(AFMT_S16_NE)
#  include <machine/endian.h>
#  if BYTE_ORDER == LITTLE_ENDIAN
#  define  AFMT_S16_NE AFMT_S16_LE
#  else
#  define  AFMT_S16_NE AFMT_S16_BE
#  endif
#  endif
#elif defined(__NetBSD__) || defined(__OpenBSD__)
#  define dsp_device "/dev/sound"
#  include <soundcard.h>
#endif

#include "pb-bonk-utility.h"

//Accuracy of fixed point calculations
const int lattice_shift = 10, sample_shift = 4, lattice_factor = 1 << lattice_shift, sample_factor = 1 << sample_shift,
//Maximum allowable taps
 max_tap = 2048;

//Default quantization level
const double base_quant = 0.6,
//Amount of bit rate variation
 rate_variation = 3.0;

const int tap_quant[max_tap] = {	//int(sqrt(i+1))
	1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4,
	4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6,
	6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
	7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
	9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10,
	10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,
	11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,
	11,11,11,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,
	12,12,12,12,12,12,12,12,13,13,13,13,13,13,13,13,13,13,13,13,
	13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,14,14,14,14,14,
	14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,
	14,14,14,14,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,
	15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,16,16,16,16,16,
	16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
	16,16,16,16,16,16,16,16,17,17,17,17,17,17,17,17,17,17,17,17,
	17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,
	17,17,17,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,
	18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,
	19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,
	19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,20,
	20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,
	20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,
	21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,
	21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,
	21,21,21,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,
	22,22,22,22,22,22,22,22,22,22,22,22
};

inline int
shift (int a, int b)
{
	return a + (1 << b - 1) >> b;
}

inline int
shift_down (int a, int b)
{
	return (a >> b) + (a < 0 ? 1 : 0);
}

struct lattice {
	int order;
	int *k, *state;

	void init (int _order) {
		order = _order;
		k = (int *)calloc (order, sizeof (int));
		state = (int *)calloc (order, sizeof (int));
	}
 
	void dequantize () {
		for (int i = 0; i < order; i++)
			k[i] *= tap_quant[i];
	}

	void init_state () {
		for (int i = order - 2; i >= 0; i--) {
			int x = state[i], new_x;
			for (int j = 0, p = i + 1; p < order; j++, p++) {
				new_x = x + shift_down (k[j] * state[p], 
							lattice_shift);
				state[p] += shift_down (k[j] * x, 
							lattice_shift);
				x = new_x;
			}
		}
	}

	/* This is now done in parallel with calculating the
	 * reflection coefficients
	 *
	 * int advance_by_value(int value) { // returns error
	 *   int  x = value,
	 *       *k_ptr = &(k[0]),
	 *       *state_ptr = &(state[0]);
	 *   for(int i=0;i<order;i++,k_ptr++,state_ptr++) {
	 *     int k_value =     *k_ptr,
	 *         state_value = *state_ptr;
	 *     state[i] += shift_down(k_value*x,lattice_shift);
	 *     x += shift_down(k_value*state_value,lattice_shift);
	 *   }
	 *
	 *   for(int i=order-1;i>0;i--)
	 *     state[i] = state[i-1];
	 *   state[0] = value;
	 * 
	 *   return x;
	 * } 
	 */

	int advance_by_error (int error) {	// returns value
		int x = error;
		x -= shift_down (k[order - 1] * state[order - 1], 
				 lattice_shift);

		int *k_ptr = &(k[order - 2]), *state_ptr = &(state[order - 2]);
		for (int i = order - 2; i >= 0; i--, k_ptr--, state_ptr--) {
			int k_value = *k_ptr, state_value = *state_ptr;
			x -= shift_down (k_value * state_value, lattice_shift);
			state_ptr[1] = state_value + shift_down 
				(k_value * x, lattice_shift);
		}

		//Don't drift too far, to avoid overflows 
		if (x > (sample_factor << 16))
			x = (sample_factor << 16);
		if (x < -(sample_factor << 16))
			x = -(sample_factor << 16);

		state[0] = x;

		return x;
	}
};

struct decoder {
	FILE *f_in;
	bitstream_in bit_in;
	int length, length_remaining, rate;
	int channels;
	bool lossless;
	bool mid_side;
	int n_taps;
	int down_sampling, samples_per_packet;
	double quant_level;

	lattice predictor;
	int **predictor_initer;

	void begin (FILE * _f_in) {
		f_in = _f_in;

		vector < char >buffer;
		for (int i = 0; i < 5; i++)
			 buffer.push_back (fgetc (f_in));

		for (;;) {
			if (buffer[buffer.size () - 5] == 0 &&
			    buffer[buffer.size () - 4] == 'B' &&
			    buffer[buffer.size () - 3] == 'O' && 
			    buffer[buffer.size () - 2] == 'N' && 
			    buffer[buffer.size () - 1] == 'K')
				break;

			buffer.push_back (fgetc (f_in));
			
			if (feof (f_in) || buffer.size () > 1000000)
				return;
		} 
		fprintf (stderr, "%s", &(buffer[0]));

		if (read_uint8 (f_in) != 0)	// version
			return;
		
		length = read_uint32 (f_in);
		rate = read_uint32 (f_in);
		channels = read_uint8 (f_in);
		lossless = read_uint8 (f_in);
		mid_side = read_uint8 (f_in);
		n_taps = read_uint16 (f_in);
		down_sampling = read_uint8 (f_in);
		samples_per_packet = read_uint16 (f_in);

		if (channels == 0 || (channels < 2 && mid_side) || 
		    n_taps > max_tap || n_taps == 0 || down_sampling == 0 || 
		    samples_per_packet == 0)
			return;

		predictor.init (n_taps);
		predictor_initer = (int **)calloc (channels, sizeof (int *));
		for (int i = 0; i < channels; i++)
			predictor_initer[i] = (int *)calloc (n_taps, sizeof (int));

		length_remaining = length;

		bit_in.setup (f_in);
	}

	void read_packet (vector < int >&samples) {
		samples.resize (samples_per_packet * down_sampling * channels);

		int input_samples[samples_per_packet];

		read_list (predictor.k, predictor.order, false, bit_in);

		predictor.dequantize ();

		int quant = (lossless ? 1 : bit_in.read_uint (16) * 
			     sample_factor);

		for (int channel = 0; channel < channels; channel++) {
			int *sample = &(samples[channel]);

			predictor.state = (int *)predictor_initer[channel];
			predictor.init_state ();
			
			read_list (input_samples, samples_per_packet, 
				       true, bit_in);
			
			for (int i = 0; i < samples_per_packet; i++) {
				for (int j = 0; j < down_sampling - 1; j++) {
					*sample = predictor.advance_by_error 
						(0);
					sample += channels;
				}

				*sample = predictor.advance_by_error 
					(input_samples[i] * quant);
				sample += channels;
			}

			for (int i = 0; i < n_taps; i++)
				predictor_initer[channel][i] = 
					samples[samples.size () - 
						channels + channel - 
						i * channels];
		}

		if (mid_side)
			for (int i = 0; i < samples.size (); i += channels) {
				samples[i + 1] += shift (samples[i], 1);
				samples[i] -= samples[i + 1];
			}

		if (!lossless)
			for (int i = 0; i < samples.size (); i++)
				samples[i] = shift (samples[i], sample_shift);

		if (length_remaining < samples.size ()) {
			samples.resize (length_remaining);
			length_remaining = 0;
		} else
			length_remaining -= samples.size ();
	}
};

FILE *
open_dsp (int rate, bool stereo)
{
	int device = open (dsp_device, O_WRONLY);
	if (device < 0)
		return NULL;

	int format = AFMT_S16_NE;
	if (ioctl (device, SNDCTL_DSP_SETFMT, &format) < 0)
		return NULL;

	int is_stereo = stereo ? 1 : 0;
	if (ioctl (device, SNDCTL_DSP_STEREO, &is_stereo) < 0)
		return NULL;

	if (ioctl (device, SNDCTL_DSP_SPEED, &rate) < 0)
		return NULL;

	return fdopen (device, "wb");
}

void
play_file (char *filename)
{
	FILE *f_out;
	decoder deco;

	FILE *f_in = fopen (filename, "rb");
	free (filename);

	if (!f_in) {
		return;
	}
	deco.begin (f_in);

	if (deco.channels > 2) {
		fclose (f_out);
		return;
	}

	f_out = open_dsp (deco.rate, deco.channels > 1);

	if (f_out == NULL) {
		fclose (f_in);
		return;
	}

	while (deco.length_remaining) {
		vector < int >samples;
		deco.read_packet (samples);
		
		vector < int16 > little_samples (samples.size ());
		for (int i = 0; i < samples.size (); i++) {
			if (samples[i] > 32767)
				little_samples[i] = 32767;
			else if (samples[i] < -32768)
				little_samples[i] = -32768;
			else
				little_samples[i] = samples[i];
		}
		fwrite (&(little_samples[0]), 2, little_samples.size (), f_out);
		fflush (f_out);
	}

	fclose (f_in);
	fclose (f_out);

	return;
}
