#include <ncurses.h>
#include <stdlib.h>
#include <string.h>

#include "ui-scroll-list.h"

static void clear_line (int y, int sx, int ex);

UIScrollList::UIScrollList (int new_start_x, int new_end_x,
			    int new_start_y, int new_end_y)
{
	start_x = new_start_x;
	end_x = new_end_x;
	start_y = new_start_y;
	end_y = new_end_y;

	list = NULL;
	list_size = 0;
}

void 
UIScrollList::set_x_y (int new_start_x, int new_end_x,
		       int new_start_y, int new_end_y)
{
	start_x = new_start_x;
	end_x = new_end_x;
	start_y = new_start_y;
	end_y = new_end_y;

	render_list ();

	if ((end_y - start_y) - 1 <= mark_position) {
		mark_position = (end_y - start_y) - 1;
		mark_position_old = mark_position;
	}

	render_mark ();
}

UIScrollList::~UIScrollList ()
{
	if (list != NULL) {
		for (int i = 0; i < list_size; i++)
			free (list[i]);
		free (list);
	}
}

void
UIScrollList::init_list_size (int new_list_size)
{
	if (list != NULL) {
		for (int i = 0; i < list_size; i++)
			free (list[i]);
		free (list);
	}

	list_size = new_list_size;	
	list = (char **)calloc (list_size, sizeof (char *));

	mark_position = 0;
	mark_position_old = 0;
	
	first_visible = 0;
}

void
UIScrollList::init_item_add (const char *item)
{
	int i;
	for (i = 0; list[i] != NULL; i++);
	list[i] = strdup (item);
}

void
UIScrollList::init_render_list (void)
{
	render_list ();
	render_mark ();
}

void
UIScrollList::render_list (void)
{
	int showed = 0;
	
	while (end_y - start_y > showed &&
	       list_size > showed) {
		clear_line (start_y + showed, start_x, end_x);
		mvaddstr (start_y + showed, start_x, 
			  list[showed + first_visible]);
		showed++;
	}
	while (showed != end_y - start_y) {
		clear_line (start_y + showed, start_x, end_x);
		showed++;
	}
}

void
UIScrollList::render_mark (void)
{
	/* Remove the old mark */
	clear_line (start_y + mark_position_old, start_x, end_x);
	attron (A_NORMAL);
	mvaddstr (start_y + mark_position_old, start_x, 
		  list[mark_position_old + first_visible]);
	attroff (A_NORMAL);
	
	/* Render the new mark */
	clear_line (start_y + mark_position, start_x, end_x);
	attron (A_STANDOUT);
	mvaddstr (start_y + mark_position, start_x, 
		  list[mark_position + first_visible]);
	attroff (A_STANDOUT);
}

void 
UIScrollList::scroll_up (void)
{
	/* Can we scroll up? */
	if (first_visible == 0 && mark_position == 0) return;

	if (mark_position == 0) {
		first_visible--;
		render_list ();
	} else {
		mark_position_old = mark_position;
		mark_position--;
	}
	render_mark ();
}

void 
UIScrollList::scroll_down (void)
{
	mark_position_old = mark_position;
	
	/* Can we scroll down? */
	if (list_size - 1 == mark_position) return;

	/* Is the mark at the bottom of the screen? */
	if ((end_y - start_y) - 1 <= mark_position) {
		/* Yes it is. Are we at the bottom of the list? */
		if (mark_position + first_visible == list_size - 1) return;
		mark_position_old--;
		first_visible++;
		render_list ();
		render_mark ();
		return;
	}
	mark_position++;
	render_mark ();
}

void 
UIScrollList::scroll_page_up (void)
{
	/* Can we scroll up? */
	if (first_visible == 0 && mark_position == 0) return;

	if (first_visible == 0) {
		mark_position_old = mark_position;
		mark_position = 0;
		render_mark ();
		return;
	}
	if (mark_position != 0) {
		mark_position_old = mark_position;
		mark_position = 0;
		render_mark ();
		return;
	}

	if (first_visible < end_y - start_y) {
		mark_position = 0;
		first_visible = 0;
		render_list ();
		render_mark ();
	} else {
		mark_position = 0;
		first_visible -= end_y - start_y;
		render_list ();
		render_mark ();
	}
}

void 
UIScrollList::scroll_page_down (void)
{
	/* Will this scroll put us in the end of the list? */
	if (first_visible + (end_y - start_y) > list_size) {
		mark_position_old = mark_position;
		mark_position = list_size - 1;
		render_mark ();
		return;
	}
	
	if (mark_position == (end_y - start_y) - 1) {
		/* Add a page to first visible */
		first_visible += end_y - start_y;
		
		/* Is first_visible to large because of this? */
		if (first_visible > list_size - (end_y - start_y))
			first_visible = list_size - (end_y - start_y);
		render_list ();
		render_mark ();
		return;
	}
	mark_position_old = mark_position;
	mark_position = end_y - start_y - 1;
	render_mark ();
}

const char *
UIScrollList::get_item (int number) const
{
	if (number >= list_size) {
		endwin ();
		printf ("item number [%i] is >= list_size [%i]\n",
			number, list_size);
		abort ();
	}
	return list[number];
}

const char *
UIScrollList::get_marked_item (void) const
{
	return list[first_visible + mark_position];
}

int
UIScrollList::get_marked_item_list_number (void) const
{
	return first_visible + mark_position;
}

static void 
clear_line (int y, int sx, int ex)
{
	for (;sx <= ex; sx++)
		mvaddch (y, sx, ' ');
}
